/* @(#) sciibin.c       2.0 03/06/89  */
/*************************************************************************
 **                                                                     **
 **  Name    :  sciibin                                                 **
 **  Author  :  Marcel J.E. Mol                                         **
 **  Date    :  06/03/89              (first release)                   **
 **  Version :  2.00                                                    **
 **  Files   :  sciibin.c       Main source file                        **
 **                                                                     **
 **  ------------------------- Revision List -------------------------  **
 **  Ver   Date       Name                   Remarks                    **
 **  1.00  06/03/89   Marcel Mol        First release                   **
 **  1.10  27/03/89   Marcel Mol        Fished things up, error routine,**
 **                                     linecount, all info fields ok,  **
 **                                     changed info layout.            **
 **  2.00  03/06/89   Marcel Mol        Bug fixes: alfabet to unsigned  **
 **                                     char. Some OS's need types.h.   **
 **                                     Output filenames can be wrong   **
 **                                     when multiple files are         **
 **                                     extracted. Made crc and file-   **
 **                                     type arrays in seperate includes**
 **  =================================================================  **
 **                                                                     **
 **  Compile as follows: cc sciibin.c -Oso sciibin                      **
 **                                                                     **
 **  Usage: sciibin [-hvtc] [-o<outputfile>] <infiles>                  **
 **                                                                     **
 **           -v show only info on file, do not create output.          **
 **           -t test file, do not create output.                       **
 **           -c do not check checksums.                                **
 **           -o create given filename instead of the one in            **
 **              binscii file. Use this only if the input files         **
 **              contain only one output file.                          **
 **           -h help.                                                  **
 **                                                                     **
 **                                                                     **
 **    Defining DEBUG gives some debug information.                     **
 **    Defining DEBUGALL gives input and output of the decode           **
 **    routine.                                                         **
 **                                                                     **
 **   This software is freeware. I can not be held responsible to       **
 **   any damage this program may coase you or anyone else.             **
 **                                                                     **
 **   Mail bugs to:      marcel@duteca.tudelft.nl                       **
 **                                                                     **
 ************************************************************************/


#include <stdio.h>
#include <sys/types.h>          /* keep sun OS happy                */
#include <sys/stat.h>           /* mostly includes types.h ...      */
#include "crc.h"
#include "filetypes.h"

#define BUFLEN 256

/*
 * Global variables
 */
char buf[BUFLEN+1];
unsigned char dec[BUFLEN+1];
unsigned char alfabet[BUFLEN+1];/* need $FF in alfabet              */
char outfilename[BUFLEN+1];

int outflag = 0;                /* the -o option                    */
int crcflag = 0;                /* the -c option                    */
int infoflag = 0;               /* the -v option                    */
int testflag = 0;               /* the -t option                    */
int makeoutput;                 /* ! (-t | -v)                      */
int crcok;                      /* -t | -c                          */
FILE *outfp;                    /* output file                      */

int namlen, filetype, numblocks; 
long filesize, startbyte, segmentlen;
int modyear, modmonth, modday, modhour, modmin;
int creyear, cremonth, creday, crehour, cremin;
int auxtype, stortype, access;
char * infilename;
int linecount;

char * progname;


/*
 * function declarations
 */
sciibin();
getheader();
decode();
decodestring();
char *myfgets();
void usage();
void error();


char * copyright = "@(#) sciibin.c  2.0 03/06/89  (c) M.J.E. Mol";

main(argc, argv)
int argc;
char **argv;
{
    FILE *fp;
    int c;
    extern int optind;                       /* For getopt */
    extern char * optarg;            /* For getopt */
    int  flag;                               /* Flag for getopt */

    progname = *argv;

    while ((flag = getopt(argc, argv, "hvcto:")) != EOF) { /* Process options */
        switch (flag) {
            case 'v': infoflag = 1;  /* Want only info of file */
                      break;
            case 'h': usage();               /* Give help */
                      exit(0);
            case 'c': crcflag = 1;
                      break;
            case 't': testflag = 1;
                      break;
            case 'o': strcpy(outfilename, optarg);
                      outflag = 1;
                      break;
            default : fprintf(stderr, "%s: skipping unkown flag %c, use -h.\n",
                                    progname, flag);
                      break;
        }
    }

    makeoutput = !(testflag | infoflag);
    crcok = testflag | crcflag;
#if defined(DEBUG)
    fprintf(stderr, "make output: %d, crcok: %d\n", makeoutput, crcok);
#endif

    if (optind >= argc) {            /* No files given, use stdin */
        infilename = "stdin";
        linecount = 0;
        sciibin(stdin);
    }
    else
        while (optind < argc) {
            infilename = argv[optind];
            optind++;
            if ((fp = fopen(infilename, "r")) == NULL) {
                perror(infilename);
                continue;
            }
            linecount = 0;
            sciibin(fp);
            fclose(fp);
        }

    exit(0);

} /* main */



/*
 * Walk over the file processing all segments in it
 */
sciibin(fp)
FILE *fp;
{
    int processed = 0;          /* number of processed binscii segments     */
    int status = 0;             /* return codes of calls to decode          */

    while (myfgets(buf, BUFLEN, fp) != NULL) {
#if defined(DEBUG)
        fprintf(stderr, "(%s) get start:%s", infilename, buf);
#endif
        if (!strcmp(buf, "FiLeStArTfIlEsTaRt\n")) {
            if (!getheader(fp) && !infoflag) /* if header ok and not -v flag */
                status |= decode(fp);
            processed++;
        }
    }

    if (processed == 0) {
        error("not a binscii file");
        return 1;
    }

    return status;

} /* sciibin */



/*
 * Build the alfabet, get the output file info and open output file
 * if necessary.
 * Still contains lots of debug code to find out the header structure.
 *   (every bit is known now...)
 */
getheader(fp)
FILE *fp;
{
    register int i, j;
    register int crc = 0;
    struct stat statbuf;            /* must know if file exists         */
    char *iomod;                    /* write or readwrite a file        */

    /*
     * Get the alfabet
     */
    if (myfgets(buf, BUFLEN, fp) == NULL) {
        error("reading alfabet: unexpected end of file");
        return 1;
    }
#if defined(DEBUG)
    fprintf(stderr, "(%s) alfabet:%s", infilename, buf);
#endif
    if (strlen(buf) != 65) {
        error("alfabet corrupted");
        return 1;
    }

    /*
     * Process the alfabet
     */
    for (i = 0; i < BUFLEN; i++)
        alfabet[i] = 0xff;
    for (i = 0; i < 64; i++) {
        j = buf[i];
        if (alfabet[j] != 0xff)
            error("Warning: double character in alfabet");
        alfabet[j] = i;
    }
#if defined(DEBUG)
    for (i = 0; i < BUFLEN; i+=16) {
        fprintf(stderr, "(%s) alfabet[%3d] =", infilename, i);
        for (j = 0; j < 16; j++)
            fprintf(stderr, " %02X", alfabet[i+j]);
        putc('\n', stderr);
    }
#endif

    /*
     * Get the file header
     */
    if (myfgets(buf, BUFLEN, fp) == NULL) {
        error("reading fileheader: unexpected end of file");
        return 1;
    }
#if defined(DEBUG)
    fprintf(stderr, "(%s) fileheader:%s", infilename, buf);
#endif

    /*
     * Extract output filename if needed
     */
    if (!outflag) {
        namlen = *buf - 'A' + 1; /* IS +1 NEEDED ?? */
        strncpy(outfilename, buf+1, namlen);
        outfilename[namlen] = '\0';
    }
#if defined(DEBUG)
    fprintf(stderr, "(%s) filename:**%s**\n", infilename, outfilename);
    fprintf(stderr, "(%s) fileinfo:**%s**", infilename, buf+16);
#endif

    /*
     * Decode and process the file header information
     */
    if ((i = decodestring(buf+16, dec)) != 27)
        error("warning: corrupted file header length");
    for (i = 0; i < 24; i++) 
        crc = updcrc(dec[i], crc);
    if (crc != (dec[24] | (dec[25] << 8))) {
        if (crcok) 
            error("warning: CRC error in file header");
        else {
            error("error: CRC error in file header");
            return 1;
        }
    }
    filesize   = dec[0] + (dec[1]<<8) + (dec[2]<<16);/* Calculate file length */
    startbyte  = dec[3] + (dec[4]<<8) + (dec[5]<<16);
    access     = dec[6];
    filetype   = dec[7];
    auxtype    = dec[8] + (dec[9] << 8);
    stortype   = dec[10];
    numblocks  = dec[11] + (dec[12]<<8);
#define CRE 13          /* perhaps creation and modification date are swapped */
    creday     = dec[CRE] & 0x1f;
    cremonth   = ((dec[CRE+1] & 0x01) << 3) | (dec[CRE] >> 5);
    creyear    = dec[CRE+1] >>1;
    cremin     = dec[CRE+2] & 0x3f;
    crehour    = dec[CRE+3] & 0x1f;
#define MOD CRE+4
    modday     = dec[MOD] & 0x1f;
    modmonth   = ((dec[MOD+1] & 0x01) << 3) | (dec[MOD] >> 5);
    modyear    = dec[MOD+1] >>1;
    modmin     = dec[MOD+2] & 0x3f;
    modhour    = dec[MOD+3] & 0x1f;
    segmentlen = dec[21] + (dec[22]<<8) + (dec[23]<<16);

#define READ    0x01
#define WRITE   0x02
#define BACKUP  0x20
#define RENAME  0x40
#define DESTROY 0x80

    if (infoflag) {
        printf("%-15s  %3s aux: $%04X ",
                   outfilename, filetypes[filetype], auxtype);
        putchar(access & READ    ? 'r' : '-');
        putchar(access & WRITE   ? 'w' : '-');
        putchar(access & RENAME  ? 'n' : '-');
        putchar(access & DESTROY ? 'd' : '-');
        putchar(access & BACKUP  ? 'b' : '-');
        switch (stortype) {
            case 0x0F : printf(" voldir"); break;
            case 0x0D : printf(" dir"); break;
            case 0x01 : printf(" seed"); break;
            case 0x02 : printf(" sap"); break;
            case 0x03 : printf(" tree"); break;
            default   : printf(" ???"); break;
        }
        printf(" %02d/%02d/%02d(%02d:%02d) -", modyear, modmonth, modday,
                                               modhour, modmin);
        printf(" %02d/%02d/%02d(%02d:%02d)\n", creyear, cremonth, creday,
                                               crehour, cremin);
        printf("Part %4d of %4d,", startbyte / 0x3000 + 1,
                                   (filesize + 0x2FFF) / 0x3000);
        printf(" bytes %7d to %7d of %7d bytes, %5d blocks\n",
                startbyte, startbyte+segmentlen, filesize, numblocks);
    }

    if (makeoutput) {
        iomod = (stat(outfilename, &statbuf) == 0) ? "r+" : "w"; 
        if ((outfp = fopen(outfilename, iomod)) == NULL) {
            error("unable to open output file");
            perror(outfilename);
            return 1;
        }
        fseek(outfp, startbyte, 0);
    }

    return 0;

} /* getheader */



/*
 * Do the actual decoding of the bin data.
 */
decode(fp)
FILE *fp;
{
    register int i;
    register int crc = 0;
    int len;

    crc = 0;
    while (segmentlen > 0) { 
        if (myfgets(buf, BUFLEN, fp) == NULL) {
            error("reading file: unexpected end of file");
            return 1;
        }
#if defined(DEBUG)
        fprintf(stderr, "(%s) data:%s", infilename, buf);
#endif
        if ((len = decodestring(buf, dec)) != 48)
            error("warning: corrupted line length");
        for (i = 0; i < 48; i++)
            crc = updcrc(dec[i], crc);
        if (makeoutput)
            for (i = 0; (i < len) && (segmentlen > 0); i++, segmentlen--)
                putc(dec[i], outfp);        /* COULD CR/NL TXT FILES */
        else
            segmentlen -= len;
#if defined(DEBUG)
        fprintf(stderr, "(%s) still need %d bytes\n", infilename, segmentlen);
#endif
    }

    /*
     * must be at end of segment now, with one remaining line containing
     * the crc check.
     */

    if (myfgets(buf, BUFLEN, fp) == NULL) {
        error("reading file crc: unexpected end of file");
        return 1;
    }
#if defined(DEBUG)
    fprintf(stderr, "(%s) crc:%s", infilename, buf);
#endif
    if ((len = decodestring(buf, dec)) != 3)
       error("warning: corrupted crc length");
    if (crc != (dec[0] | (dec[1] << 8))) {
        if (crcok) 
            error("warning: CRC error in file data");
        else {
            error("error: CRC error in file data");
            return 1;
        }
    }

    fclose(outfp);

    return 0;

} /* decode */


/*
 * Decode one string off scii characters to binary data, meanwhile
 * calculating crc.
 */
decodestring(in, out)
register char *in;
register unsigned char *out;
{
    register int len = 0;

#if defined(DEBUGALL)
    char *b;

    fprintf(stderr, "(%s) decode in: %s\n", infilename, in);
    b = in;
    while (*b) fprintf(stderr, ".%02X", alfabet[*b++]);
    putc('\n', stderr);
    b = out;
#endif
    while (strlen(in) > 3) {
        *out++ = ((alfabet[in[3]] << 2) | (alfabet[in[2]] >> 4)) & 0xFF;
        *out++ = ((alfabet[in[2]] << 4) | (alfabet[in[1]] >> 2)) & 0xFF;
        *out++ = ((alfabet[in[1]] << 6) | (alfabet[in[0]]))      & 0xFF;
        len += 3;
        in += 4;
    }
    *out = '\0';
    if (*in != '\0' && *in != '\n') 
        error("warning: line not ended by NULL or NEWLINE");
#if defined(DEBUGALL)
    fprintf(stderr, "(%s) decode out:\n", infilename);
    while (b != out) 
        fprintf(stderr, ".%02X", *b++);
    putc('\n', stderr);
#endif

    return len;

} /* decodestring */



char *myfgets(buf, len, fp)
char *buf;
int len;
FILE *fp;
{

    linecount++;
    return fgets(buf, len, fp);

} /* myfgets */



void usage()
{

    fprintf(stderr, "Usage: sciibin [-vtc] [-o<outputfile>] <infiles>\n\n");
    fprintf(stderr, "       -v show only info on file, do not create output.\n");
    fprintf(stderr, "       -t test file, do not create output.\n");
    fprintf(stderr, "       -c do not check checksums.\n");
    fprintf(stderr, "       -o create given filename instead of the one in\n");
    fprintf(stderr, "          binscii file. Use this only if the input files\n");
    fprintf(stderr, "          contain only one output file.\n");
    fprintf(stderr, "       -h this help message.\n");

} /* usage */



void error(str)
char *str;
{

    fprintf(stderr, "%s (%s, %d): %s\n", 
                    progname, infilename, linecount, str);

} /* error */
